#!/usr/bin/perl
#
#
#		 Naiocast - Yet Another MP3 Streaming Server
#
#		Copyleft (C) 2001-2008 / THEARTOFWEB Software
#
# Naiocast is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# Naiocast is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Naiocast; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
#
# Author:
# Giangaetano Vituzzi - THEARTOFWEB Software
#
# You may contact the author of Naiocast by e-mail at:
# naiocast@theartofweb.net
#
# The latest version of Naiocast can be obtained from:
# http://www.naiocast.theartofweb.net
#
# v1.8.4 / receiver.cgi / Stable Release
#
#

use strict;
use warnings;
use Socket;
use File::stat;
use IO::Handle '_IONBF';
use Fcntl ':flock';
use constant BUFSIZE => 1024 * 2;

open STDERR, ">>", ".log.txt";

##
 # If we get a "GET" call: from a browser
##

if ( $ENV{"REQUEST_METHOD"} eq 'GET')
{
	print "Content-type: text/plain\n\n";
	if (! -e "online.txt")
	{
		print "Naiocast Server - Status: Active (OK)\n";
	} else {
		print "Naiocast Server - Status: Broadcasting / Up Time: " . &_get_uptime() . " - " . scalar(&_getlist()) . " Listening\n";
	}
	exit;
}

##
 # If we get a "HEAD" call: from naioSC
##

if ( $ENV{"REQUEST_METHOD"} eq 'HEAD' )
{
	if($ENV{"HTTP_USER_AGENT"})
	{
	  	if(&_check_header_password($ENV{"HTTP_USER_AGENT"}))
	  	{
	  		print "Status: 200 OK\n\n";
	  	} else {
	  		print "Status: 401 Wrong Password\n\n";
	  	}
	}
	exit;
}

##
 # If we don't get a "POST" fires 400 Bad Request
##

if( $ENV{"REQUEST_METHOD"} ne 'POST' )
{
	close(STDIN);
	print "Status: 400 Bad Request\n\n";
	exit;
}

##
 # If we get here it means naioSC is sending data by "POST"
 # We may want to check $ENV{"HTTP_USER_AGENT"} again:
##

if(!&_check_header_password($ENV{"HTTP_USER_AGENT"}))
{
	close(STDIN);
	print "Status: 401 Wrong Password\n\n";
	exit;
}

##
 # Tidy things up
##

&_remove_users();

##
 # setup UDP socket for sending
##

my ($usock,$uport); ($usock,$uport) = &_udp_setup("127.0.0.1");

##
 # NaioSC doesn't send custom http header anylonger,
 # because some proxies may remove unknown key/value pairs.
 # Instead, It sends a binary header on top of the raw data.
 # The binary header will look something like this:
 #
 # NAIO $xx $xx $xx $xx KEY1 $xx $xx $xx $xx VALUE1 KEY2 $xx $xx $xx $xx VALUE2
 #
 # - The first 4 bytes of the header are always "NAIO"
 # - The Header size is encoded with 4 bytes (int32)
 # - Key ID is 4 chars long made out of the characters capital A-Z and 0-9
 # - Value size is encoded with 4 bytes (int32)
 # - Value can contain any kind of data
##

my $io = new IO::Handle;

my ($header, $buf);
my @listeners = &_getlist();
my $now = time;
my $tnew;

if ( $io->fdopen(fileno(STDIN),"r") )
{	
	&_notify_online();

	$io->read($buf, 4); # GET NAIO
	$header .= $buf;
	
	$io->read($buf, 4); # GET 4 Bytes
	$header .= $buf;
		
	$io->read($buf, unpack("N", $buf)); # Get the rest of the header
	$header .= $buf;
	
	my $decoded_h = &_dmap_to_hash_ref($header);
	my %h;

	$h{"Name"}     = unpack("a*", $decoded_h->{NAIO}->{NAME});
	$h{"Genre"}    = unpack("a*", $decoded_h->{NAIO}->{GENR});
	$h{"Url"}      = unpack("a*", $decoded_h->{NAIO}->{URL1});
	$h{"Bitrate"}  = unpack("n",  $decoded_h->{NAIO}->{BITR});
	$h{"Pub"}      = unpack("c",  $decoded_h->{NAIO}->{PUBB});
	$h{"Host"}     = unpack("a*", $decoded_h->{NAIO}->{HOST});
	$h{"MaxL"}     = unpack("n",  $decoded_h->{NAIO}->{MAXL});
	$h{"IcyM"}     = unpack("n",  $decoded_h->{NAIO}->{ICYM});
	$h{"Header"}   = unpack("a*", $decoded_h->{NAIO}->{HEAD});
	$h{"IDEv2"}    = unpack("a*", $decoded_h->{NAIO}->{IDE2});
		
	{open my $fh1, ">", "header.txt" or die "header.txt $!\n";
	
	my $xml_h = [
		    '<header>',
		      '<name>' . $h{"Name"}    . '</name>',
		     '<genre>' . $h{"Genre"}   . '</genre>',
		       '<url>' . $h{"Url"}     . '</url>',
		       '<pub>' . $h{"Pub"}     . '</pub>',
		   '<bitrate>' . $h{"Bitrate"} . '</bitrate>',
		       '<max>' . $h{"MaxL"}    . '</max>',
		'<icymetaint>' . $h{"IcyM"}    . '</icymetaint>',
		      '<host>' . $h{"Host"}    . '</host>',
		   '</header>'
		];
	
	print $fh1 join "\n", @$xml_h;
	close $fh1;}
	
	{open my $fh1, ">", "id3v2.txt" or die "id3v2.txt $!\n";
	binmode $fh1;
	print $fh1 $h{"IDEv2"};
	close $fh1;}
	
	{open my $fh1, ">", "ncs_header.txt" or die "ncs_header.txt $!\n";
	binmode $fh1;
	print $fh1 $h{"Header"};
	close $fh1;}
	
	##
	 # Get stream (raw data)
	##
	
	my $c = 0;
		
	while( my $len = $io->read($buf, BUFSIZE) )
	{
		##
		 # Get Song Title
		##
		
		$c += $len;
			
		if( ($c + BUFSIZE) >= $h{"IcyM"} && $h{"IcyM"} > 0)
		{
			$io->read(my $buf2, ($h{"IcyM"} - $c));
			$buf .= $buf2;
			$c = 0;
			
			$io->read(my $lmeta, 1);
			
			if($lmeta and unpack("c", $lmeta) > 0)
			{
				$io->read(my $meta, (unpack("c", $lmeta) * 16));
				
				{open my $fh1, ">", "_songtit.txt";
				&_lock($fh1);
				print $fh1 $lmeta;
				print $fh1 $meta;
				&_unlock($fh1);
				close $fh1;}
			}
		}
			
		$tnew = time();
		
		@listeners = &_getlist() and $now = $tnew if (($tnew - $now) > 0);
				
		##
		 # Send stream to listeners
		##
		
		if (scalar(@listeners) > 0)
		{
			foreach my $listener (@listeners)
			{
				chomp($listener);
				&_udp_send($usock, "127.0.0.1", $listener, $buf);
			}
		}
	}
	
	close $usock;
	$io->close;
	
	&_notify_offline();
	&_quit_clients();
	
	unlink "_songtit.txt";
	unlink "ncs_header.txt";
	unlink "header.txt";
	unlink "id3v2.txt";
}

&_remove_users();

print "Status: 200 OK\n\n";

exit;

##
 # Subroutines
##

sub _dmap_to_hash_ref {
    my $buf = shift;
    my %tags;

    while (length $buf) {
        my ($tag, $len) = unpack("a4N", $buf);
        if (!defined($len) or length $buf < 8+$len) {
            return;
          }
        my $data = substr($buf, 8, $len);
        # try to unpack--if we can, assume it was a container
        my $data2 = _dmap_to_hash_ref($data);
        $tags{$tag} = $data2 ? $data2 : $data;
        substr($buf, 0, 8+$len) = '';
      }
    return \%tags;
}

sub _remove_users()
{
	my @users = <*.user>;
	my $num = @users;
	if ($num > 0)
	{
		foreach my $user (@users)
		{
			chomp($user);
			unlink ($user);
		}
	}
}

sub _notify_online()
{
	qx(touch online.txt);
}

sub _notify_offline()
{
	unlink("online.txt");
}

sub _quit_clients()
{
	my @pids = &_getpids();
	my $num = @pids;
	if ($num > 0)
	{
		foreach my $pid (@pids)
		{
			qx(kill -3 $pid >/dev/null 2>&1);
		}
	}
}

sub _getlist {

	my @empty;
	my @users = <*.user>;
	my $num = @users;
	
	# Get active clients
	#
	
	if ($num > 0)
	{

		my @active_users; # Active users
		my $sb;           # File::Stat method
		my $lastmod;      # Last mod time
		my $tnew = time;  # Now time
		my $timeout = 10; # timeout seconds
		
		my (@ports, $port);

		foreach my $user (@users)
		{
			chomp($user);
				$sb = stat($user);
					$lastmod = $sb->mtime;
	
			if ( ($tnew - $lastmod) < $timeout )
			{
				push @active_users, $user;
			} else {
				unlink $user;
			}
		}
	
		foreach my $active_user (@active_users)
		{
			($port, undef, undef) = split("_", $active_user, 3);
			push @ports, $port;
		}
		
		return @ports;
		
	} else {

		return @empty;

	}

}

sub _getpids {

	my @empty;
	my @users = <*.user>;
	my $num = @users;
	my @pids;
	my $pid;

	if ($num > 0)
	{
		foreach my $user (@users)
		{
			(undef, $pid, undef) = split("_", $user, 3);
			push @pids, $pid;
		}
		
		return @pids;
		
	} else {
	
		return @empty;
		
	}

}

sub _udp_setup {
		
	my $HOSTNAME = $_[0];
	my $fh;
	
	socket($fh, PF_INET, SOCK_DGRAM, getprotobyname("udp")) or return "";
	binmode($fh);
	
	my $ipaddr   = inet_aton($HOSTNAME);
	my $portaddr = sockaddr_in(0, $ipaddr);

	bind($fh, $portaddr) or return "";
	
	my ($nport, $naddr);
	($nport, $naddr) = sockaddr_in(getsockname($fh));
	
   	return ($fh,$nport);
}

sub _udp_send {
	
	my $USOCK	 = $_[0];
	my $HOSTNAME = $_[1];
	my $PORTNO   = $_[2];
	my $MSG	     = $_[3];
	
	my $ipaddr   = inet_aton($HOSTNAME);
	my $portaddr = sockaddr_in($PORTNO, $ipaddr);

	send($USOCK, $MSG, 0, $portaddr) == length($MSG) or return 1;

	return 0;
}

sub _check_header_password
{
	my $pwd       = shift;
	my $pwdf      = '_pswd.txt';
	my $pwdh;
	($pwdh)       = ($pwd =~ /<pwd>(.*?)<\/pwd>/sg);
	$pwdh         =~ s/[^ -~]+//g;
	
	if (length($pwdh) == 0) { return 0 };
	
	if (-e $pwdf)
	{
		open my $fh1, "<", $pwdf;
		my $pwds = <$fh1>;
		close $fh1;
		
		if (substr(crypt($pwdh, substr($pwdh, 0, 2)), 2, 11) eq $pwds) { return 1 } else { return 0 };
	}
	
	open my $fh1, ">", $pwdf;
	print $fh1 substr(crypt($pwdh, substr($pwdh, 0, 2)), 2, 11);
	close $fh1;
	
	chmod 0640, $pwdf;
	
	return 1;
}

sub _get_uptime
{
	my $file     = "online.txt";
	my $sb       = stat $file;
	my $lmod     = $sb->mtime;
	my $tnow     = time;

	my $seconds = $tnow - $lmod;

	my $s = $seconds % 60;
	my $m = (($seconds - $s) % (60*60))/60;
	my $h = ($seconds - $m*60 - $s) / (60*60);

	return sprintf( "%1d Hours, %1d Minutes, %1d Seconds", $h, $m, $s );

}

sub _lock
{
	flock( $_[0], LOCK_EX );
}

sub _unlock
{
	flock( $_[0], LOCK_UN );
}