#!/usr/bin/perl
#
#
#		 Naiocast - Yet Another MP3 Streaming Server
#
#		Copyleft (C) 2001-2008 / THEARTOFWEB Software
#
# Naiocast is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# Naiocast is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Naiocast; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
#
# Author:
# Giangaetano Vituzzi - THEARTOFWEB Software
#
# You may contact the author of Naiocast by e-mail at:
# naiocast@theartofweb.net
#
# The latest version of Naiocast can be obtained from:
# http://www.naiocast.theartofweb.net
#
# v1.8.4 / naioSC.pl / Stable Release
#
#

use strict;
use warnings;
use lib "lib";
use URI;
use LWP::UserAgent;
use MP3::Tag;
use IO::Socket::INET;
use Config::INI::Simple;
use constant BUFSIZE => 1024 * 2;

##
 # Set Internal Vars
##

our $Types;
my %Type_To_Unpack;
my $Container_Type;
my $VERSION = 'Naiocast/1.8.4';
my $CRLF = "\015\012";
&_init(); # Load dmap values

##
 # Catch Signals
##

my $OS = $^O;

if ($OS =~ /^MSWin/i) {
  $OS = 'WINDOWS';
} elsif ($OS =~ /^VMS/i) {
  $OS = 'VMS';
} elsif ($OS =~ /^dos/i) {
  $OS = 'DOS';
} elsif ($OS =~ /^MacOS/i) {
    $OS = 'MACINTOSH';
} elsif ($OS =~ /^os2/i) {
    $OS = 'OS2';
} elsif ($OS =~ /^epoc/i) {
    $OS = 'EPOC';
} elsif ($OS =~ /^cygwin/i) {
    $OS = 'CYGWIN';
} else {
    $OS = 'UNIX';
}

if ($OS ne 'WINDOWS')
{
	$SIG{"PIPE"} = $SIG{"KILL"} = $SIG{"QUIT"} = $SIG{"ABRT"} = $SIG{"INT"} = sub { print "-> Connection closed!\n\n"; exit; };
}

##
 # Read Config File
##

my $conf = new Config::INI::Simple; $conf->read("config.txt");
my %settings;

$settings{"local_addr"}      = $conf->{"default"}->{"Nicecast-Addr"};
$settings{"local_port"}      = $conf->{"default"}->{"Nicecast-Port"};
$settings{"ncs_pass"}        = $conf->{"default"}->{"NCServer-Pass"};
$settings{"ncs_addr"}        = $conf->{"default"}->{"NCServer-Addr"};
$settings{"ncs_type"}        = $conf->{"default"}->{"NCServer-Header-Type"};
$settings{"ncs_mlis"}        = $conf->{"default"}->{"NCServer-Max-Listeners"};
$settings{"tagid3_file"}     = '/tmp/naiocast_id3v2.txt';
$settings{"naioheader_file"} = '/tmp/naiocast_header.txt';

my %ncs;
$ncs{"url"}    = new URI( $settings{"ncs_addr"} );
$ncs{"host"}   = $ncs{"url"}->host;
$ncs{"port"}   = $ncs{"url"}->port || 80;

##
 # Print Splash Screen (!)
##

print q|
       THEARTOFWEB Software 2001-2008

      (                              )
     (  (         NAIOCAST         )  )
    (  (  (                      )  )  )
   (  (  (                        )  )  )
  (  (  (  naioSC - Sender Client  )  )  )
   (  (  (                        )  )  )
    (  (  (                      )  )  )
     (  (       Version 1.8.4      )  )
      (                              )

|;
print "-> Running on: $OS ($^O)\n\n";

##
 # Check Naiocast Server
##

print qq|-> Checking Naiocast Server at $ncs{"host"} on port $ncs{"port"} ...\n|;

my $browser = LWP::UserAgent->new;

$browser->timeout(5);
$browser->agent(qq|Mozilla/4.76 [en] (Windows NT 5.0; U) <pwd>$settings{"ncs_pass"}</pwd>|);
$browser->parse_head(0);

my $ncs_resp = $browser->head($settings{"ncs_addr"});

if($ncs_resp->code != 200)
{
	print "--> Naiocast Server Error! (" . $ncs_resp->message . ")\n"; exit;
} else {
	print "-> Response: Naiocast Server is Active\n";
}

##
 # Check Audio Source and connect
##

print qq|\n-> Checking Audio Source at $settings{"local_addr"} on port $settings{"local_port"} ...\n|;

my $gsock = &_sock_connect($settings{"local_addr"}, $settings{"local_port"});
   $gsock->autoflush(1);

my $h = [
	'GET / HTTP/1.1',
	'Accept: */*',
	'User-Agent: RMA/1.0 (' . $VERSION . ')',
	'Icy-MetaData: 1',
	'SupportsMaximumASMBandwidth: 1',
	'Connection: Keep-Alive',
	'Accept-Language: en-us',
'',''
];

syswrite $gsock, join $CRLF, @$h;

my %sr;
$sr{"status"} = <$gsock>;
$sr{"status"} =~ s/\r?\n$//;

($sr{"protocol"}, $sr{"code"}, $sr{"message"}) = split(' ', $sr{"status"}, 3);

if ($sr{"code"} != 200)
{
	close $gsock;
	die qq|--> naioSC Error! $settings{"local_addr"} on port $settings{"local_port"} MAY NOT be streaming ($sr{"code"}) $sr{"message"} \n|;
}

my @h; push @h, $sr{"status"};

while (<$gsock>)
{
	s/\r?\n$//;
	s/[^ -~]+//g;
	last unless $_;
	push @h, $_;
}

my %h;

foreach (@h)
{
	my ($key, $value) = split(":", $_, 2);
	if($value) {
		$value =~ s/^\s+//;
		$h{$key} = $value;
	}
}
push @h, '', '';

##
 #   $h{"icy-notice1"};
 #   $h{"icy-notice2"};
 # * $h{"icy-name"};
 # * $h{"icy-genre"};
 # * $h{"icy-url"};
 # * $h{"icy-pub"};
 # * $h{"icy-br"};
 # * $h{"icy-metaint"};
 #   $h{"Content-Type"};
 #   $h{"Server"};
##

$h{"icy-name"}      = 'undef name' unless $h{"icy-name"};
$h{"icy-genre"}     = 'undef genre' unless $h{"icy-genre"};
$h{"icy-url"}       = 'undef url' unless $h{"icy-url"};
$h{"icy-pub"}       = 0 unless $h{"icy-pub"};
$h{"icy-br"}        = 0 unless $h{"icy-br"};
$h{"icy-metaint"}   = 0 unless $h{"icy-metaint"};

my $asres; # Audio Source Response
	
     if ( $sr{"protocol"} =~ m/ICY/ ) {
     
		$h{"icy-notice2"} =~ s/<BR>//i;
		$asres = qq|-> Response: $h{"icy-notice2"}\n\n|;
		
} elsif ( $sr{"protocol"} =~ m/HTTP/ ) {

		$asres = qq|-> Response: $h{"Server"}\n\n|;
		
} else {

		$asres = qq|--> Response: Unknown Audio Source Generator!\n\n|;
}

##
 # Create SHOUTcast or Icecast Header
##

my $ncs_header;

     if ($settings{"ncs_type"} == 1) { # SHOUTcast

	$_ = [
		'ICY 200 OK',
		'icy-notice1:<BR>This stream requires <a href="http://www.winamp.com/">Winamp</a><BR>',
		'icy-notice2:SHOUTcast Distributed Network Audio Server/Linux v1.9.8<BR>',
		'icy-name:'    . $h{"icy-name"},
		'icy-genre:'   . $h{"icy-genre"},
		'icy-url:'     . $h{"icy-url"},
		'icy-pub:'     . $h{"icy-pub"},
		'icy-br:'      . $h{"icy-br"},
		'icy-metaint:' . $h{"icy-metaint"},
		'content-type:audio/mpeg',
	'',''
	];
	$ncs_header = join $CRLF, @$_;
	
} elsif ($settings{"ncs_type"} == 2) { # Icecast
	$_ = [
		'HTTP/1.0 200 OK',
		'Content-Type: audio/mpeg',
		'icy-br:'      . $h{"icy-br"},
		'icy-genre:'   . $h{"icy-genre"},
		'icy-name:'    . $h{"icy-name"},
		'icy-pub:'     . $h{"icy-pub"},
		'icy-url:'     . $h{"icy-url"},
		'icy-metaint:' . $h{"icy-metaint"},
		'Server: Icecast 2.3.1',
	'',''
	];
	$ncs_header = join $CRLF, @$_;
	
} else {
	die qq|--> naioSC Error! Unknown Header type: $settings{"ncs_type"}\n\n|;
}

print $asres;

##
 # Create ID3v2.3.0 Header
##
	
{
	open my $fh1, ">", $settings{"tagid3_file"} or die "$!\n";
	close $fh1;
}

my $_mp3 = MP3::Tag->new($settings{"tagid3_file"});
my $_id3v2 = $_mp3->new_tag("ID3v2");

$_id3v2->add_frame('TIT2', $h{"icy-name"});
$_id3v2->add_frame('TIT3', "Powered by $VERSION");
$_id3v2->add_frame('TPE1', $h{"icy-genre"});
$_id3v2->add_frame('COMM', "ENG", "Short text", "Powered By $VERSION");
$_id3v2->add_frame('TEXT', "Powered By $VERSION");
$_id3v2->add_frame('WOAR', $h{"icy-url"});
$_id3v2->add_frame('WORS', $h{"icy-url"});
$_id3v2->add_frame('TCON', "Web Radio");
$_id3v2->write_tag();

my $_id3v2_h = do { open my $fh1, '<', $settings{"tagid3_file"} or die; binmode $fh1; undef $/; <$fh1> };

##
 # Create Naiocast Binary Header
##
	
my $_nbh = &_dmap_pack([
[ 'Naiocast' => [
	[ 'Name'        => $h{"icy-name"}           ],
	[ 'Genre'       => $h{"icy-genre"}          ],
	[ 'Url'         => $h{"icy-url"}            ],
	[ 'Bitrate'     => $h{"icy-br"}             ],
	[ 'Pub'         => $h{"icy-pub"}            ],
	[ 'Host'        => 'http://' . $ncs{"host"} ],
	[ 'MaxL'        => $settings{"ncs_mlis"}    ],
	[ 'IcyM'        => $h{"icy-metaint"}        ],
	[ 'Header'      => $ncs_header              ],
	[ 'IDEv2'       => $_id3v2_h                ]
]]]);

{
	open my $fh1, ">", $settings{"naioheader_file"} or die "$!\n";
	binmode $fh1;
	print $fh1 $_nbh;
	close $fh1;
}

##
 # Send the stream to Naiocast Server
##

print qq|-> Preparing for Broadcasting ...\n|;

my $naio_header     = do { open my $fh1, '<', $settings{"naioheader_file"} or die; binmode $fh1; undef $/; <$fh1> };
my $naio_headerl    = do { use bytes; length $naio_header };
	
my $ua = LWP::UserAgent->new;
$ua->agent(qq|Mozilla/4.76 [en] (Windows NT 5.0; U) <pwd>$settings{"ncs_pass"}</pwd>|);

my $req = HTTP::Request->new(POST => $settings{"ncs_addr"});

$req->content_type('application/octet-stream');
$req->content_length(2048000000);

print qq|-> Sending Naiocast Binary Header ($naio_headerl Bytes) ... Sent!\n|;
print qq|-> Relaying from $settings{"local_addr"} to $ncs{"host"} ... \n|;
print qq|-> Broadcasting ... (you are on air!)\n|;
my $_hsent = 0;

$req->content(sub {
	if($_hsent == 0)
	{
		$_hsent = 1;
		return $naio_header;
	} else {
		read($gsock, my $buf, BUFSIZE);
		return $buf;
	}
});

my $res = $ua->request($req);

close $gsock;

print "-> Connection Closed! " . $res->code . "\n\n";

exit;

##
 # Internal Subroutines
##

sub _sock_connect {

	my $addr = shift;
	my $port = shift;
	my $sock = IO::Socket::INET->new(PeerAddr => $addr, 
										PeerPort => $port, 
											Proto => 'tcp', 
												Timeout => 9) or 
													die "--> naioSC Error: $addr on port $port (Connection Refused)\n";
	return $sock;
}

##
 # I would like to thank Richard Clamp for the following code:
##

sub _init {
    local $/;
    $Types = eval <DATA>;

    $Container_Type = 12;

    %Type_To_Unpack = (
                       1 => 'c',
                       3 => 'n',
                       5 => 'N',
                       7 => 'Q',
                       9 => 'a*',  # utf-8 encoded
                       10 => 'N',
                       11 => 'nn',
                       );
}

sub _dmap_pack {
    my $struct = shift;
    my $out = '';

    my %by_name = map { $_->{NAME} => $_ } values %$Types;
    for my $pair (@$struct) {
        my ($name, $value) = @$pair;        
        my $tag  = $by_name{ $name }{ID};
        my $type = $by_name{ $name }{TYPE};
        
        if ($type == 12) { # container
            $value = _dmap_pack( $value );
        }
        elsif ($type == 7) { # 64-bit
            my $high = Math::BigInt->new( $value )->brsft(32)."";
            my $low  = Math::BigInt->new( $value )->band(0xFFFFFFFF)."";
            $value = pack( "N2", $high, $low );
        }
        else {
            no warnings 'uninitialized';
            $value = pack( $Type_To_Unpack{$type}, $value );
        }
        my $length = do { use bytes; length $value };
        $out .= $tag . pack("N", $length) . $value;
    }
    return $out;
}

__DATA__
        {
          'NAIO' => {
                      'ID' => 'NAIO',
                      'NAME' => 'Naiocast',
                      'TYPE' => 12
                    },
          'NAME' => {
                      'ID' => 'NAME',
                      'NAME' => 'Name',
                      'TYPE' => 9
                    },
          'GENR' => {
                      'ID' => 'GENR',
                      'NAME' => 'Genre',
                      'TYPE' => 9
                    },
          'URL1' => {
                      'ID' => 'URL1',
                      'NAME' => 'Url',
                      'TYPE' => 9
                    },
          'HOST' => {
                      'ID' => 'HOST',
                      'NAME' => 'Host',
                      'TYPE' => 9
                    },
          'BITR' => {
                      'ID' => 'BITR',
                      'NAME' => 'Bitrate',
                      'TYPE' => 3
                    },
          'MAXL' => {
                      'ID' => 'MAXL',
                      'NAME' => 'MaxL',
                      'TYPE' => 3
                    },
          'ICYM' => {
                      'ID' => 'ICYM',
                      'NAME' => 'IcyM',
                      'TYPE' => 3
                    },
          'PUBB' => {
                      'ID' => 'PUBB',
                      'NAME' => 'Pub',
                      'TYPE' => 1
                    },
          'HEAD' => {
                      'ID' => 'HEAD',
                      'NAME' => 'Header',
                      'TYPE' => 9
                    },
          'IDE2' => {
                      'ID' => 'IDE2',
                      'NAME' => 'IDEv2',
                      'TYPE' => 9
                    }
        }